# Simple interface to some SLURM commands
#
# Copyright (c) 2017 Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

import pandas as pd
import re
import StringIO
import subprocess
import types

from pygcam.mcs.error import PygcamMcsException
from pygcam.log import getLogger

_logger = getLogger(__name__)

class Slurm(object):
    def __init__(self, *args, **kwargs):
        """
        Create a Slurm instance for interacting with SLURM commands.

        :param defaults: (dict) default values for common parameters so
            they don't have to be passed to all the individual commands.
        """
        import getpass

        self.username = kwargs.get('username') or getpass.getuser()

    def squeue(self, formatStr='%all', sep='|', otherArgs=''):
        """
        Run the squeue command and return a DataFrame with the results.

        :param formatStr: (str) format string indicating which data columns
           to list. Default is to list all columns, separated by '|'.
        :param sep: (str) separator to use between data columns
        :param otherArgs (str) other args to pass to squeue
        :return: status of shell command
        """
        command = "squeue -u %s -o '%s' %s" % (self.username, formatStr, otherArgs)
        _logger.debug(command)

        # Run command and read all results
        output = subprocess.check_output(command, shell=True)

        df = pd.read_table(StringIO.StringIO(output), sep=sep, header=0, engine='c')
        df.fillna('', inplace=True)

        # Parse the overloaded nodelist column into separate columns
        colName = 'NODELIST(REASON)'
        if colName in df.columns:
            values = df[colName]
            pat1 = re.compile('([^\(]*)')       # anything up to a '('
            pat2 = re.compile('(\(.+\))') # anything between '(' and ')'

            def search(pat, s):
                matchObj = re.search(pat, s)
                return matchObj.group(1) if matchObj else ''

            df['NODELIST'] = map(lambda s: search(pat1, s), values)
            df['REASON']   = map(lambda s: search(pat2, s), values)

        return df

    def jobsInState(self, state, jobName=None):     # TBD: test this again
        """
        Return a list of jobIds for jobs in the given state.

        :param state: (str) standard SLURM state string (PENDING, RUNNING, etc.)
        :return: (list of int) the ids of the jobs owned by the user in the given
            state.
        """
        import platform
        if platform.system() != 'Linux':
            return []

        df = self.squeue(formatStr='%i|%j', otherArgs='-t ' + state.upper())
        jobs = df.loc[df.NAME == jobName, 'JOBID'] if jobName else df.JOBID
        return jobs

    def scancel(self, jobs):
        """
        Terminate the jobs identified by the given jobIds.

        :param jobs: (str or iterable of str) a jobId or sequence of jobIds
        :return: none
        """
        if isinstance(jobs, (list, tuple, set, types.GeneratorType)):
            jobStr = ' '.join(jobs)
        else:
            jobStr = str(jobs)

        command = "scancel " + jobStr
        _logger.debug(command)

        exitStatus = subprocess.call(command, shell=True)
        if exitStatus != 0:
            raise PygcamMcsException("Command failed: %s; exit status %s\n" % (command, exitStatus))

    def sbatch(self, script, queue):
        """
        Submit the given script to the given queue.

        :return: (int) jobId or -1 if shell command failed
        """
        command = "sbatch %s" % (script)

        # Run the sbatch command, parse the jobId if command succeeds
        jobStr = subprocess.check_output(command, shell=True)
        result = re.search('\d+', jobStr)
        jobId = int(result.group(0)) if result else -1
        return jobId


if __name__ == "__main__":
    output = '''JOBID|PARTITION|NAME|STATE|TIME|NODES|NODELIST(REASON)|TIMELIMIT|USER
3515574|shared|14100-5|RUNNING|1-23:35:42|1|node122|7-00:00:00|nand374
3515573|shared|14100-4|RUNNING|1-23:36:26|1|node122|7-00:00:00|nand374
3515572|shared|14100-3|RUNNING|1-23:37:09|1|node122|7-00:00:00|nand374
3515568|shared|14100-2|RUNNING|1-23:42:01|1|node122|7-00:00:00|nand374
3515438|shared|t14_4_1|PENDING|2-01:08:35|1||7-00:00:00|nand374
3515439|shared|t14_4_2|PENDING|2-01:08:35|1||7-00:00:00|nand374
3515440|shared|t14_4_3|RUNNING|2-01:08:35|1|node141|7-00:00:00|nand374
3515441|shared|t14_4_4|RUNNING|2-01:08:35|1|node141|7-00:00:00|nand374
3515442|shared|t14_4_5|RUNNING|2-01:08:35|1|node141|7-00:00:00|nand374
3515341|shared|t14_4_5|RUNNING|2-03:07:55|1|node141|7-00:00:00|nand374
3515339|shared|t14_4_4|RUNNING|2-03:11:47|1|node141|7-00:00:00|nand374
3515336|shared|t14_4_3|RUNNING|2-03:14:31|1|node141|7-00:00:00|nand374
3515334|shared|t14_4_2|RUNNING|2-03:15:30|1|node141|7-00:00:00|nand374
3515333|shared|t14_4_1|RUNNING|2-03:15:50|1|node141|7-00:00:00|nand374
'''
    pd.set_option('display.width', 1000)
    # df = pd.read_table(StringIO.StringIO(output), sep='|', header=0, engine='c')

    slurm = Slurm(username='plev920')
    df = slurm.squeue(formatStr="%i|%P|%j|%u|%T|%M|%D|%R|%l")
    print(df)

    jobs = slurm.jobsInState('pending')
    print("Pending: %s" % jobs)

    jobs = slurm.jobsInState('running')
    print("Running: %s" % jobs)




# Created on 5/11/15
#
# Copyright (c) 2015-2017. The Regents of the University of California (Regents).
# See the file COPYRIGHT.txt for details.
import os
from collections import OrderedDict
from datetime import datetime

import pandas as pd

from ..config import getParam
from ..log import getLogger
from ..XMLFile import XMLFile
from .error import PygcamMcsUserError, PygcamMcsSystemError, FileMissingError
from .Database import getDatabase
from .XML import XMLWrapper, findAndSave

_logger = getLogger(__name__)

RESULT_TYPE_DIFF = 'diff'
RESULT_TYPE_SCENARIO = 'scenario'
DEFAULT_RESULT_TYPE = RESULT_TYPE_SCENARIO

QUERY_OUTPUT_DIR = 'queryResults'

RESULT_ELT_NAME = 'Result'
FILE_ELT_NAME = 'File'
CONSTRAINT_ELT_NAME = 'Constraint'
COLUMN_ELT_NAME = 'Column'

class XMLConstraint(XMLWrapper):
    equal = ['==', '=', 'equal', 'eq']
    notEqual = ['!=', '<>', 'notEqual', 'not equal', 'neq']

    def __init__(self, element):
        super(XMLConstraint, self).__init__(element)
        self.column = element.get('column')
        self.op = element.get('op')
        self.value = element.get('value')
        self.groupBy = element.get('groupby')

        if self.op:
            known = self.equal + self.notEqual
            if not self.op in known:
                raise PygcamMcsUserError('Unknown operator in constraint: %s' % self.op)

            if not self.value:
                raise PygcamMcsUserError('Constraint with operator "%s" is missing a value' % self.op)

    def asString(self):
        op = '==' if self.op in self.equal else ('!=' if self.op in self.notEqual else None)

        if op:
            return "%s %s %r" % (self.column, op, self.value)

        return None


class XMLColumn(XMLWrapper):
    def __init__(self, element):
        super(XMLColumn, self).__init__(element)


class XMLResult(XMLWrapper):
    '''
    Represents a single Result (model output) from the results.xml file.
    '''
    def __init__(self, element):
        super(XMLResult, self).__init__(element)
        self.name = element.get('name')
        self.type = element.get('type', DEFAULT_RESULT_TYPE)
        self.desc = element.get('desc')
        self.queryFile = self._getPath(FILE_ELT_NAME)

        col = self.element.find(COLUMN_ELT_NAME)
        self.column = XMLColumn(col) if col is not None else None

        # Create the "where" clause to use with a DataFrame.query() on the results we'll read in
        self.constraints = map(XMLConstraint, self.element.iterfind(CONSTRAINT_ELT_NAME))
        constraintStrings = filter(None, map(XMLConstraint.asString, self.constraints))
        self.whereClause = ' and '.join(constraintStrings)

    def isScalar(self):
        return self.column is not None

    def _getPath(self, eltName):
        'Get a single filename from the named element'
        objs = self.element.findall(eltName)
        filename = objs[0].get('name')
        if os.path.isabs(filename):
            raise PygcamMcsUserError("For %s named %s: path (%s) must be relative" % (eltName, self.name, filename))

        return filename

    def csvPathname(self, scenario, baseline=None, outputDir='.', type=RESULT_TYPE_SCENARIO):
        """
        Compute the pathname of a .csv file from an outputDir,
        scenario name, and optional baseline name.
        """
        # Output files are stored in the output dir with same name as query file but with 'csv' extension.
        basename = os.path.basename(self.queryFile)
        mainPart, extension = os.path.splitext(basename)
        middle =  scenario if type == RESULT_TYPE_SCENARIO else ("%s-%s" % (scenario, baseline))
        csvFile = "%s-%s.csv" % (mainPart, middle)
        csvPath = os.path.abspath(os.path.join(outputDir, csvFile))
        return csvPath

    def columnName(self):
        return self.column.getName() if self.column is not None else None


class XMLResultFile(XMLFile):
    """
    XMLResultFile manipulation class.
    """

    cache = {}

    @classmethod
    def getInstance(cls, filename):
        try:
            return cls.cache[filename]

        except KeyError:
            obj = cls.cache[filename] = cls(filename)
            return obj

    def __init__(self, filename):
        super(XMLResultFile, self).__init__(filename, load=True, schemaPath='mcs/etc/results-schema.xsd')
        root = self.tree.getroot()

        self.results = OrderedDict()    # the parsed fileNodes, keyed by filename
        findAndSave(root, RESULT_ELT_NAME, XMLResult, self.results)

    def getResultDefs(self, type=None):
        """
        Get results of type 'diff' or 'scenario'
        """
        results = self.results.values()

        if type:
            results = filter(lambda result: result.type == type, results)

        return results

    def saveOutputDefs(self):
        '''
        Save the defined outputs in the SQL database
        '''
        db = getDatabase()
        session = db.Session()
        for result in self.getResultDefs():
            db.createOutput(result.name, description=result.desc, session=session)

        session.commit()
        db.endSession(session)

    @classmethod
    def addOutputs(cls):
        resultsFile = getParam('MCS.ResultsFile')
        obj = cls(resultsFile)
        obj.saveOutputDefs()

class QueryResult(object):
    '''
    Holds the results of an XPath batch query
    '''
    def __init__(self, filename):
        self.filename = filename
        self.title = None
        self.df    = None
        self.units = None
        self.readCSV()

    @staticmethod
    def parseScenarioString(scenStr):
        """
        Parse a GCAM scenario string into a name and a time stamp

        :param scenStr: (str) a scenario name string of the form
           "Reference,date=2014-29-11T08:10:45-08:00"

        :return: (str, datetime) the scenario name and a datetime instance
        """
        name, datePart = scenStr.split(',')
        # _logger.debug("datePart: %s", datePart)
        dateWithTZ = datePart.split('=')[1]     # drop the 'date=' part
        # _logger.debug("dateWithTZ: %s", dateWithTZ)

        # drops the timezone info with strptime doesn't handle.
        # TBD: this is ok as long as all the scenarios were run in the same timezone...
        lenTZ = len("-00:00")
        dateWithoutTZ = dateWithTZ[:-lenTZ]
        # _logger.debug("dateWithoutTZ: %s", dateWithoutTZ)

        runDate = datetime.strptime(dateWithoutTZ, "%Y-%d-%mT%H:%M:%S")   # N.B. order is DD-MM, not MM-DD
        return name, runDate

    def readCSV(self):
        '''
        Read a CSV file produced by a batch query. The first line is the name of the query;
        the second line provides the column headings; all subsequent lines are data. Data
        are comma-delimited, and strings with spaces are double-quoted. Assume units are
        the same as in the first row of data.
        '''
        _logger.debug("readCSV: reading %s", self.filename)
        with open(self.filename) as f:
            self.title  = f.readline().strip()
            self.df = pd.read_table(f, sep=',', header=0, index_col=False, quoting=0)

        df = self.df

        if 'Units' in df.columns:
            self.units = df.Units[0]

        # split the scenario field into two parts; here we create the columns
        df['ScenarioName'] = None
        df['ScenarioDate'] = None

        if 'scenario' in df.columns:        # not the case for "diff" files
            name, date = self.parseScenarioString(df.loc[0].scenario)
            df['ScenarioName'] = name
            df['ScenarioDate'] = date

    def getFilename(self):
        return self.filename

    def getTitle(self):
        return self.title

    def getData(self):
        return self.df


def collectResults(context, type):
    '''
    Called by worker to process results, return a list of dicts
    with data the master process can quickly write to the database.
    '''
    from collections import defaultdict
    from .util import getSimResultFile, activeYears, YEAR_COL_PREFIX

    _logger.debug("Collecting results for %s", context)

    baseline = context.baseline
    scenario = context.scenario

    if type == RESULT_TYPE_DIFF:
        assert baseline, "saveResults: must specify baseline for DIFF results"

    resultsFile = getSimResultFile(context.simId)
    rf = XMLResultFile.getInstance(resultsFile)
    outputDefs = rf.getResultDefs(type=type)

    if not outputDefs:
        _logger.info('saveResults: No outputs defined for type %s', type)
        return []

    # TBD: was used to delete prior results; may not be needed
    # names = map(XMLResult.getName, outputDefs)

    activeYears = activeYears()
    yearCols = map(lambda y: YEAR_COL_PREFIX + y, activeYears)

    # A single result DF can have data for multiple outputs, so we cache the files
    outputCache = defaultdict(lambda: None)

    trialDir = context.getTrialDir()

    # TBD: create context.getQueryResultsDir() ?
    scenarioOutputDir = os.path.join(trialDir, scenario, 'queryResults')

    # TBD: create context.getDiffsDir() ?
    diffsOutputDir = os.path.join(trialDir, scenario, 'diffs')

    outputDir = scenarioOutputDir if type == RESULT_TYPE_SCENARIO else diffsOutputDir

    resultList = []

    for output in outputDefs:
        csvPath = output.csvPathname(scenario, baseline=baseline, outputDir=outputDir, type=type)

        if not outputCache[csvPath]:
            try:
                outputCache[csvPath] = QueryResult(csvPath)
            except Exception as e:
                _logger.warning('saveResults: Failed to read query result: %s', e)
                raise FileMissingError(csvPath)

        queryResult = outputCache[csvPath]
        paramName   = output.name
        whereClause = output.whereClause

        selected = queryResult.df.query(whereClause) if whereClause else queryResult.df
        count = selected.shape[0]
        if count == 0:
            raise PygcamMcsUserError('Query where clause(%r) matched no results' % whereClause)

        if 'region' in selected.columns:
            firstRegion = selected.region.iloc[0]
            if count == 1:
                regionName = firstRegion
            else:
                _logger.debug(
                    "Query where clause (%r) yielded %d rows; year columns will be summed" % (whereClause, count))
                regionName = firstRegion if len(selected.region.unique()) == 1 else 'Multiple'
        else:
            regionName = 'global'

        # Create a dict to return. No need to store runId or scenario; context has these already
        resultDict = dict(regionName=regionName, paramName=paramName)

        if output.isScalar():
            colName = output.columnName()
            value = selected[colName].iloc[0]
            resultDict['isScalar'] = True
        else:
            # When no column name is specified, assume this is a time-series result, so save all years.
            # Use sum() to collapse values to a single time series; for a single row it helpfully
            # converts the 1-element series to a simple float.
            value = {colName: selected[yearStr].sum() for colName, yearStr in zip(yearCols, activeYears)}
            resultDict['isScalar'] = False
            resultDict['units'] = queryResult.units

        resultDict['value'] = value
        resultList.append(resultDict)

    # a list of dicts with results for this trial
    return resultList

def saveResults(context, resultList):
    '''
    Called on the master to save results to the database that were prepared by the worker.
    '''
    from .Database import getDatabase

    runId = context.runId

    db = getDatabase()
    session = db.Session()

    # Delete any stale results for this runId (i.e., if re-running a given runId)
    names = map(lambda resultDict: resultDict['paramName'], resultList)
    ids = db.getOutputIds(names)
    db.deleteRunResults(runId, outputIds=ids, session=session)
    db.commitWithRetry(session)

    for resultDict in resultList:
        paramName  = resultDict['paramName']
        value      = resultDict['value']
        regionName = resultDict['regionName']
        regionId = db.getRegionId(regionName)

        # Save the values to the database
        try:
            if resultDict['isScalar']:
                db.setOutValue(runId, paramName, value, session=session)  # TBD: need regionId?
            else:
                units = resultDict['units']
                db.saveTimeSeries(runId, regionId, paramName, value, units=units, session=session)

        except Exception as e:
            session.rollback()
            db.endSession(session)
            # TBD: distinguish database save errors from data access errors?
            raise PygcamMcsSystemError("saveResults failed: %s" % e)

    db.commitWithRetry(session)
    db.endSession(session)



"""
.. Support for generating land-protection scenarios described in an XML file.

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from __future__ import print_function
import copy
import os
import sys

from lxml import etree as ET

from .config import getParam
from .constants import UnmanagedLandClasses, GCAM_32_REGIONS
from .error import FileFormatError, CommandlineError, PygcamException
from .log import getLogger
from .utils import mkdirs, pathjoin, flatten, parse_version_info
from .XMLFile import XMLFile

_logger = getLogger(__name__)

Verbose = False

def pp(elt):
    print(ET.tostring(elt, pretty_print=True))

def _findChildren(node, tag, cls=None):
    """
    Find all the children beneath `node` with the given `tag`, and
    return an instance of the class `cls` representing that node.

    :param node: the node to find children from
    :param tag: the tag for the children to find
    :param cls: optional class to instantiate for each child. If
      not specified, the class is assumed to be named by the
      capitalized version of the tag, i.e., tag 'foo' implies class Foo.
      As a special case, if cls == str, a list of the text contents for
      the given tag is returned.
    :return: a list of elements of type 'cls' or the imputed class.
    """

    if not cls:
        className = tag[0].upper() + tag[1:]    # retains camelCase
        thisModule = sys.modules[__name__]
        cls = getattr(thisModule, className)

    nodes = node.findall(tag)
    if cls == str:
        children = map(lambda node: node.text, nodes)
    else:
        children = map(cls, nodes)

    return children


class LandProtection(object):
    """
    Stores the application's representation of the parsed XML file describing land
    protection scenarios.

    :param node: an ``lxml.etree.Element`` representing the top-level ``<landProtection>`` node
    """
    def __init__(self, node):
        self.groups    = _findChildren(node, 'group')
        self.scenarios = _findChildren(node, 'scenario')

    def getScenario(self, name):
        return Scenario.getScenario(name)

    def protectLandTree(self, tree, scenarioName):
        """
        Apply the protection scenario `scenarioName` to the parsed XML file `tree`.
        This interface is provided so WriteFuncs (which are passed an open XMLInputFile)
        can apply protection scenarios.

        :param tree: (lxml ElementTree) a tree for a parsed XML input file.
        :param scenarioName: (str) the name of the scenario to apply
        :return: none
        """
        _logger.info("Applying protection scenario %s", scenarioName)

        scenario = Scenario.getScenario(scenarioName)
        if not scenario:
            raise FileFormatError("Scenario '%s' was not found" % scenarioName)

        # Iterate over all definitions for this scenario, applying the protections
        # incrementally to the tree representing the XML file that was read in.
        for protReg in scenario.protRegDict.values():
            regions = [protReg.name]
            for prot in protReg.protections:
                createProtected(tree, prot.fraction, landClasses=prot.landClasses, regions=regions)

    # TBD: test this
    def protectLand(self, infile, outfile, scenarioName, backup=True, unprotectFirst=False):
        """
        Generate a copy of `infile` with land protected according to `scenarioName`,
        writing the output to `outfile`.

        :param infile: input file (should be one of the GCAM aglu-xml land files)
        :param outfile: the file to create which represents the desired land protection
        :param scenarioName: a scenario in the landProtection.xml file
        :param backup: if True, create a backup `outfile`, with a '~' appended to the name,
          before writing a new file.
        :param unprotectFirst: (bool) if True, make all land "unprotected" before protecting.
        :return: none
        """
        parser = ET.XMLParser(remove_blank_text=True)
        tree = ET.parse(infile, parser)

        # TBD: eliminate this for v5.0
        # Remove any existing land protection, if so requested
        if unprotectFirst:
            unProtectLand(tree, otherArable=True)

        self.protectLandTree(tree, scenarioName)

        if backup:
            try:
                # Ensure we're not clobbering reference files.
                backupFile = outfile + '~'
                os.rename(outfile, backupFile)
            except Exception as e:
                PygcamException('Failed to create backup file "%s": %s', backupFile, e)

        _logger.info("Writing '%s'...", outfile)
        tree.write(outfile, xml_declaration=True, pretty_print=True)


class Group(object):
    Instances = {}

    def __init__(self, node):
        self.regions = _findChildren(node, 'region', cls=str)
        self.Instances[node.get('name')] = self

    @classmethod
    def globalGroup(cls):
        '''
        Generate built-in 'global' region
        '''
        group = ET.Element('group', name='Global')
        for name in GCAM_32_REGIONS:
            reg = ET.SubElement(group, 'region')
            reg.text = name

        # print(ET.tostring(group, pretty_print=True))
        return cls(group)

    @classmethod
    def getGroup(cls, name):
        group = cls.Instances.get(name)

        if not group and name.lower() == 'global':
            group = cls.globalGroup()

        return group

    @classmethod
    def regionNames(cls, groupName):
        group = cls.getGroup(groupName)
        return map(lambda region: region.name, group.regions)

    @staticmethod
    def expandNames(name, names):
        '''
        Recursively expand a list of region/group names into a list of actual region names.
        This is generally not called directly; it is called from ProtectedRegion, which is
        where the expansion is required.

        :param name: the name of the initial region/group to expand
        :param names: a list which holds the accumulated region names
        :return: nothing -- the result is in the list `names` passed by the user.
        '''
        group = Group.getGroup(name)
        if not group:
            names.append(name)
        else:
            for region in group.regions:
                Group.expandNames(region, names)

class ProtectedRegion(object):
    def __init__(self, node):
        self.name = node.get('name')
        self.protections = _findChildren(node, 'protection')

    def expandNames(self):
        names = []
        Group.expandNames(self.name, names)
        return names

class Protection(object):
    def __init__(self, node):
        self.fraction = float(node.find('fraction').text)
        self.landClasses = _findChildren(node, 'landClass', cls=str)

class Scenario(object):
    Instances = {}

    def __init__(self, node):
        self.name = node.get('name')
        self.protRegs = _findChildren(node, 'protectedRegion')
        self.Instances[self.name] = self
        self.protRegDict = protRegDict = {}       # allows regions to override groups

        for protReg in self.protRegs:
            _logger.debug("Defining protectedRegion '%s'", protReg.name)
            regions = protReg.expandNames()

            # print(protReg.name, ' => ', regions)
            for region in regions:
                regCopy = copy.copy(protReg)
                regCopy.name = region
                protRegDict[region] = regCopy

    @classmethod
    def getScenario(cls, name):
        return cls.Instances.get(name)

def _makeRegionXpath(regions):
    if not regions:
        return ""

    if isinstance(regions, (str, unicode)):
        regions = [regions]

    patterns = map(lambda s: "@name='%s'" % s, regions)
    regionPattern = ' or '.join(patterns)
    xpath = "//region[%s]" % regionPattern
    _logger.debug('regionXpath: ' + xpath)
    return xpath

def _makeLandClassXpath(landClasses, protected=False):
    if isinstance(landClasses, (str, unicode)):
        landClasses = [landClasses]

    prefix = 'Protected' if protected else ''
    patterns = map(lambda s: 'starts-with(@name, "%s%s")' % (prefix, s), landClasses)
    landPattern = ' or '.join(patterns)
    xpath = ".//UnmanagedLandLeaf[%s]" % landPattern
    _logger.debug('landClassXpath: ' + xpath)
    return xpath

def unProtectLand(tree, landClasses=None, otherArable=False, regions=None):
    """
    Restore the file to 0% land protection by adding the protected land back
    into its unprotected counterpart and deleting the protected elements.

    :param tree: a tree representing a parsed GCAM land_input XML file
    :param landClasses: a string or a list of strings, or None. If None, all
           standard unmanaged land classes are modified.
    :param otherArable: (bool) if True, land class 'OtherArableLand' is
        included in default land classes.
    :param regions: a string or a list of strings, or None. If None, all
           regions are modified.
    :return: None
    """
    regionXpath = _makeRegionXpath(regions) if regions else ''
    landRoots = tree.xpath(regionXpath + '//LandAllocatorRoot')

    if not landClasses:
        landClasses = UnmanagedLandClasses + (['OtherArableLand'] if otherArable else [])

    protectedXpath = _makeLandClassXpath(landClasses, protected=True)

    for landRoot in landRoots:
        protectedNodes = landRoot.xpath(protectedXpath)

        if len(protectedNodes) == 0:
            continue

        # Find matching not-protected node and add protected land back in
        for node in protectedNodes:
            name = node.get('name')
            unProtectedName = name[len("Protected"):]
            prefix = './/UnmanagedLandLeaf[@name="%s"]' % unProtectedName

            protectedAllocs = node.xpath(".//allocation|.//landAllocation")

            for alloc in protectedAllocs:
                year = alloc.get('year')
                xpath = prefix + '//%s[@year="%s"]' % (alloc.tag, year)
                unprotectedAlloc = landRoot.find(xpath)
                originalArea = float(unprotectedAlloc.text) + float(alloc.text)
                unprotectedAlloc.text = str(originalArea)

        # Remove all the protected nodes, restoring the file to its original state
        landNodes = landRoot.xpath('./LandNode[starts-with(@name, "Protected")]')
        for landNode in landNodes:
            parent = landNode.getparent()
            parent.remove(landNode)

def createProtected(tree, fraction, landClasses=None, otherArable=False,
                    regions=None, unprotectFirst=False):
    """
    Modify an lxml tree representing a GCAM input file to protect a `fraction`
    of `landClasses` in `regions`.

    :param tree: a tree representing a parsed GCAM land_input XML file
    :param fraction: the fraction of land in the given land classes to protect
    :param landClasses: a string or a list of strings, or None. If None, all
           standard unmanaged land classes are modified.
    :param otherArable: (bool) if True, land class 'OtherArableLand' is
        included in default land classes.
    :param regions: a string or a list of strings, or None. If None, all
           regions are modified.
    :param unprotectFirst: (bool) if True, make all land "unprotected" before
           protecting.
    :return: None
    """
    version = parse_version_info()
    if version >= (5, 0):
        raise PygcamException("Called landProtection.createProtected on GCAM version >= 5.0. Use landProtectionUpdate.protectLand instead.")

    _logger.debug('createProtected: fraction=%.2f, landClasses=%s, regions=%s, unprotect=%s',
                  fraction, landClasses, regions, unprotectFirst)

    def multiplyValues(nodes, factor):
        for n in nodes:
            newValue = float(n.text) * factor
            n.text = str(newValue)

    # Remove any existing land protection, if so requested
    if unprotectFirst:
        unProtectLand(tree, landClasses=landClasses, otherArable=otherArable, regions=regions)

    regionXpath = (_makeRegionXpath(regions) if regions else '') + '//LandAllocatorRoot'
    landRoots = tree.xpath(regionXpath)

    if not landClasses:
        landClasses = UnmanagedLandClasses + (['OtherArableLand'] if otherArable else [])

    unmgdXpath     = _makeLandClassXpath(landClasses)
    protectedXpath = _makeLandClassXpath(landClasses, protected=True)

    allocXpath = ".//allocation|.//landAllocation"

    for landRoot in landRoots:
        # ensure that we're not protecting an already-protected land class in these regions
        nodes = landRoot.xpath(protectedXpath)
        if len(nodes) > 0:
            node = nodes[0]
            regNodes = list(node.iterancestors(tag='region'))
            region = regNodes[0].get('name')
            raise FileFormatError('Error: Land class %s is already protected in region %s' % (node.tag, region))

        nodes = landRoot.xpath(unmgdXpath)

        for node in nodes:
            landnode = ET.SubElement(landRoot, 'LandNode')

            new = copy.deepcopy(node)
            newName = 'Protected' + node.get('name')
            new.set('name', newName)
            landnode.set('name', newName)
            landnode.set('fraction', "%.4f" %fraction)
            landnode.append(new)

            originalAreas = node.xpath(allocXpath)
            protectedAreas = new.xpath(allocXpath)

            multiplyValues(originalAreas, 1 - fraction)
            multiplyValues(protectedAreas, fraction)

def protectLand(infile, outfile, fraction, landClasses=None, otherArable=False,
                regions=None, unprotectFirst=False):
    """
    Create a copy of `infile` that protects a `fraction` of `landClasses` in `regions`.

    :param infile: the path of a GCAM land_input XML file
    :param outfile: the path of the XML file to create by modifying data from `infile`
    :param fraction: the fraction of land in the given land classes to protect
    :param landClasses: a string or a list of strings, or None. If None, all
        "standard" unmanaged land classes are modified.
    :param otherArable: (bool) if True, land class 'OtherArableLand' is
        included in default land classes.
    :param regions: a string or a list of strings, or None. If None, all
        regions are modified.
    :param unprotectFirst: (bool) if True, make all land "unprotected" before
        protecting.
    :return: None
    """
    parser = ET.XMLParser(remove_blank_text=True)
    tree = ET.parse(infile, parser)

    createProtected(tree, fraction, landClasses=landClasses, otherArable=otherArable,
                    regions=regions, unprotectFirst=unprotectFirst)
    tree.write(outfile, xml_declaration=True, pretty_print=True)

_LandXmlFiles = ['land2.xml', 'land3.xml']

def _landXmlPaths(workspace, landXmlFiles=_LandXmlFiles):
    xmlDir = pathjoin(workspace, 'input', 'gcam-data-system', 'xml', 'aglu-xml')
    paths = map(lambda filename: pathjoin(xmlDir, filename), landXmlFiles)
    return paths

def parseLandProtectionFile(scenarioFile=None):
    scenarioFile = scenarioFile or getParam('GCAM.LandProtectionXmlFile')
    xmlFile = XMLFile(scenarioFile, schemaPath='etc/protection-schema.xsd')
    obj = LandProtection(xmlFile.getRoot())
    return obj

def runProtectionScenario(scenarioName, outputDir=None, workspace=None,
                          scenarioFile=None, xmlFiles=None, inPlace=False,
                          unprotectFirst=False):
    """
    Run the protection named by `scenarioName`, found in `scenarioFile` if given,
    or the value of config variable `GCAM.LandProtectionXmlFile` otherwise. The source
    files are take from `workspace`, if given, otherwise from the value of `GCAM.RefWorkspace`.
    Results are written to the given `outputDir`. In the even that the input and output
    files are the same, `inPlace` must be set to True to indicate that overwriting is intended.
    By default the two files `xmlFiles`, land2.xml and land3.xml in the aglu-xml directory,
    are processed, though other files can be specified in the unlikely case that you have
    alternatives.

    :param scenarioName: (str) the name of a protection scenario defined in the `scenarioFile`
    :param outputDir: (str) the directory under which to write the modified land files. Ignored
       if inPlace=True.
    :param workspace: (str) the location of the workspace holding the input files (ignored
       if xmlFiles are specified explicitly)
    :param scenarioFile: (str) the path to a protection.xml file defining `scenarioName`
    :param xmlFiles: (list of str) the paths of the XML input files to modify
    :param inPlace: (bool) if True, input and output files may be the same (output overwrites input).
    :param unprotectFirst: (bool) if True, make all land "unprotected" before
           protecting.
    :return: none
    """
    _logger.debug("Land-protection scenario '%s'", scenarioName)

    landProtection = parseLandProtectionFile(scenarioFile=scenarioFile)

    workspace = workspace or getParam('GCAM.SandboxRefWorkspace')
    xmlFiles = xmlFiles or _landXmlPaths(workspace)

    for inFile in xmlFiles:
        basename = os.path.basename(inFile)
        outFile = inFile if inPlace else pathjoin(outputDir, basename)

        # check that we're not clobbering the input file
        if not inPlace and os.path.lexists(outFile) and os.path.samefile(inFile, outFile):
            raise CommandlineError("Attempted to overwrite '%s' but --inPlace was not specified." % inFile)

        landProtection.protectLand(inFile, outFile, scenarioName, unprotectFirst=unprotectFirst)

def protectLandMain(args):

    global Verbose
    Verbose = args.verbose
    regions      = args.regions and flatten(map(lambda s: s.split(','), args.regions))
    scenarioFile = args.scenarioFile or getParam('GCAM.LandProtectionXmlFile')
    scenarioName = args.scenario
    outDir    = args.outDir
    workspace = args.workspace or getParam('GCAM.RefWorkspace')
    template  = args.template

    if not workspace:
        raise CommandlineError('Workspace must be identified in command-line or config variable GCAM.RefWorkspace')

    if args.mkdir:
        mkdirs(outDir)

    xmlFiles = _landXmlPaths(workspace)

    # Process instructions from protection XML file
    if scenarioName:
        if not scenarioFile:
            raise CommandlineError('Scenario "%s" was specified, but a scenario file was not identified',
                                   scenarioName)
        runProtectionScenario(scenarioName, outDir, workspace=workspace,
                              scenarioFile=scenarioFile, xmlFiles=xmlFiles, inPlace=args.inPlace)
        return

    # If no scenario name given, process command-line args
    fraction = args.fraction
    if fraction is None:
        raise CommandlineError('If not using protection scenarios, fraction must be provided')

    landClasses = flatten(map(lambda s: s.split(','), args.landClasses)) if args.landClasses \
                    else UnmanagedLandClasses + (['OtherArableLand'] if args.otherArable else [])

    fraction = float(fraction)
    templateDict = {'fraction' : str(int(fraction * 100)),
                    'regions'  : '-'.join(regions) if regions else 'global',
                    'classes'  : '-'.join(landClasses) if args.landClasses else 'unmanaged'}

    for path in xmlFiles:
        filename = os.path.basename(path)
        templateDict['filename'] = filename
        templateDict['basename'] = os.path.splitext(filename)[0]

        outFile = template.format(**templateDict)
        outPath = pathjoin(outDir, outFile)
        _logger.debug("protectLand(%s, %s, %0.2f, %s, %s)", path, outFile, fraction, landClasses, regions)
        protectLand(path, outPath, fraction, landClasses=landClasses, regions=regions)



'''
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from abc import ABCMeta, abstractmethod

# class OptionInfo(object):
#     """
#     Stores information about a single sub-command option that is used to
#     construct the web-based GUI for the sub-command.
#     """
#     def __init__(self, display=True, choiceFunc=None, multiple=False):
#         """
#         Initialize an OptionInfo instance.
#
#         :param display: (bool) whether this option should be displayed in the GUI
#         :param choiceFunc: (callable) function to return a list of options to display
#         :param multiple: (bool) whether multiple values can be returned
#         :param commaList: (bool) whether values should be returned as a comma-delimited string
#         """
#         self.display = display
#         self.choiceFunc = choiceFunc
#         self.multiple = multiple

class SubcommandABC(object):
    """
    Abstract base class for sub-commands. Defines the protocol expected by ``gt``
    for defining sub-commands. Plugin files should be named ``'*_plugin.py'`` and must
    define a subclass of ``SubcommandABC``. To allow the class to be identified, it can be
    named ``Plugin`` or the global variable ``PluginClass`` can identify the class.

    :param name: (str) the name of the sub-command
    :param subparsers: an object returned by argparse's ``parser.add_subparsers()``
    :param kwargs: (dict) keywords to pass to the the call to argparse's
       ``subparsers.add_parser(name, **kwargs)``, e.g., to pass `help` or
       `documentation` strings.
    :param group: (str) the name of the GUI group to assign this plug-in to.
    :param label: (str) the label to use on the GUI to reference this plug-in.
       Defaults to the sub-command name, capitalized.
    :param guiSuppress: (bool) if True, do not display this sub-command in the GUI.
    """
    __metaclass__ = ABCMeta

    Instances = {}  # SubCommand instances keyed by name
    Parsers = {}    # SubCommand parsers keyed by name

    @classmethod
    def getInstance(cls, name):
        return SubcommandABC.Instances.get(name)

    # Deprecated?
    @classmethod
    def getParser(cls, name):
        obj = cls.getInstance(name)
        return obj.parser

    def __init__(self, name, subparsers, kwargs, group=None, label=None,
                 guiSuppress=False):
        self.name = name
        self.label = label or name.capitalize()  # label to display in GUI
        self.parser = parser = subparsers.add_parser(self.name, **kwargs)
        self.Instances[self.name] = self

        self.optionInfo = {}    # OptionInfo instances keyed by option name
        self.guiSuppress = guiSuppress

        # For grouping commands in the GUI. Set this in subclass' addArgs().
        # Set to None if the command should not be presented in the GUI.
        # Note: 'group' can be set as standalone var or in guiInfo['group'].
        self.group = group or 'main'
        #self.group = group or (guiInfo.get('group') if guiInfo else 'main')

        self.addArgs(parser)

    def __str__(self):
        clsName = type(self).__name__
        return "<%s name=%s group=%s>" % (clsName, self.name, self.group)

    def getOptionInfo(self, name):
        return self.optionInfo.get(name)

    def setOptionInfo(self, name, optionInfo):
        self.optionInfo[name] = optionInfo

    def getGroup(self):
        return self.group

    @abstractmethod
    def addArgs(self, parser):
        """
        Add command-line arguments to the given `parser`. (This is an
        abstract method that must be implemented in the subclass.)

        :param parser: the sub-parser associated with this sub-command.

        :return: the populated parser
        """
        pass

    @abstractmethod
    def run(self, args, tool):
        """
        Perform the function intended by the ``SubcommandABC`` subclass. This function
        is invoked by ``gt`` on the ``SubcommandABC`` instance whose name matches the
        given sub-command. (This is an  abstract method that must be implemented in
        the subclass.)

        :param args: the argument dictionary
        :param tool: the GcamTool instance for the main command
        :return: nothing
        """
        pass



'''
Created on Dec 19, 2012


@author: Richard Plevin
@author: Sam Fendell

Copyright (c) 2012-2015. The Regents of the University of California (Regents)
and Richard Plevin. See the file COPYRIGHT.txt for details.
'''

# Status codes for invoked programs
RUNNER_SUCCESS = 0
RUNNER_FAILURE = -1

# DistGenerate
DEFAULT_BINS     = 30
DEFAULT_TRUNCATE = 3
COUNT_TITLE      = 'count'

DISCRETE_ENTRY_SEPARATOR = ':'
DICT_SEPARATOR = '='
DISTRO_SUFFIX  = 'distro'

# SymbolicSet, Distro:
DICT_REGEX = '^[^%s]+%s[^%s]+$' % (DICT_SEPARATOR, DICT_SEPARATOR, DICT_SEPARATOR)

# Multiple users
COMMENT_CHAR = '#'
PARAM_SUFFIX = 'param'


# Default region map for GCAM 4.
# TBD: Take this from pygcam.
RegionMap = {
    'Multiple': -1,
    'global': 0,
    'USA': 1,
    'Africa_Eastern': 2,
    'Africa_Northern': 3,
    'Africa_Southern': 4,
    'Africa_Western': 5,
    'Australia_NZ': 6,
    'Brazil': 7,
    'Canada': 8,
    'Central America and Caribbean': 9,
    'Central Asia': 10,
    'China': 11,
    'EU-12': 12,
    'EU-15': 13,
    'Europe_Eastern': 14,
    'Europe_Non_EU': 15,
    'European Free Trade Association': 16,
    'India': 17,
    'Indonesia': 18,
    'Japan': 19,
    'Mexico': 20,
    'Middle East': 21,
    'Pakistan': 22,
    'Russia': 23,
    'South Africa': 24,
    'South America_Northern': 25,
    'South America_Southern': 26,
    'South Asia': 27,
    'South Korea': 28,
    'Southeast Asia': 29,
    'Taiwan': 30,
    'Argentina': 31,
    'Colombia': 32
}



"""
.. Logging support.
   This module allows modules to register themselves for logging which is
   turned on after the app reads configuration information. Modules call
   logger = pygcam.log.getLogger(__name__) as a top-level statement, evaluated
   at load time. This returns the logger, which may not yet be configured.
   When the configuration file has been read, all registered loggers are
   initialized, and all subsequently registered loggers are initialized
   upon instantiation.

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
import os
import logging
from .config import getParam, getParamAsBoolean, configLoaded

PKGNAME = __name__.split('.')[0]

_Loggers    = {}      # loggers created herein, keyed by module or package name
_LogLevels  = None    # log levels keyed by module or package name
_verbose    = False   # whether debug msgs should print

# Can't use this module to debug itself
def _debug(msg):
    if _verbose:
        print(msg)

# Note: 'traitlets' library uses root logger, which we don't want to enable here
def _createPkgLogger(dotspec):
    pkgName = dotspec.split('.')[0]

    if pkgName and pkgName not in _Loggers:
        _debug('_createPkgLogger("%s") from %s' % (pkgName, dotspec))
        logger = getLogger(pkgName)
        logger.propagate = False

def getLogger(name):
    '''
    Register a logger, which will be set up after the configuration
    file is read.

    :param name: the name of the logger, conventionally passed as __name__.
    :return: a logging logger instance
    '''
    _debug('getLogger("%s")' % name)

    try:
        logger = _Loggers[name]

    except KeyError:
        logger = logging.getLogger(name)
        logger.propagate = True             # set to False for explicitly named modules
        _Loggers[name] = logger

    _configureLogger(name)
    _createPkgLogger(name)

    return logger

def parseLevels(levelStr=None):
    """
    Get log levels for pygcam as a whole or for indicated modules individually.
    Modules starting with a '.' are interpreted to be in pygcam, i.e., ".config"
    is equivalent to "pygcam.config".
    Example: LogLevel = WARNING, .tool:DEBUG, .utils:INFO, .mcs.util:INFO, my_plugin:DEBUG

    :param levelStr: a comma-delimited string of module:logLevel values. If
        no ':' is present, the value is treated as the default logLevel for pygcam.
        If levelStr is None, the value of the variable 'GCAM.LogLevel' is used.
    :return: (dict) of log levels, keyed by module names
    """
    def splitAndStrip(s, delim):
        items = map(str.strip, map(str, s.split(delim)))
        return items

    result = {}

    levelStr = levelStr or getParam('GCAM.LogLevel')

    levels = splitAndStrip(levelStr, ',')
    for level in levels:
        if ':' in level:
            module, lvl = splitAndStrip(level, ':')
            # if '.' not in module or module[0] == '.':
            #     module = PKGNAME + '.' + (module[1:] if module[0] == '.' else module)
            if module[0] == '.':
                module = PKGNAME + module
        else:
            module = PKGNAME
            lvl = level.strip()

        result[module] = lvl.upper()

    return result

#
# Copied here from utils.py to avoid an import loop
#
def _mkdirs(newdir, mode=0o770):
    """
    Try to create the full path `newdir` and ignore the error if it already exists.
    """
    from errno import EEXIST

    try:
        os.makedirs(newdir, mode)
    except OSError as e:
        if e.errno != EEXIST:
            raise

def _addHandler(logger, formatStr, logFile=None):
    if logFile:
        _mkdirs(os.path.dirname(logFile))

    handler = logging.FileHandler(logFile, mode='a') if logFile else logging.StreamHandler()
    handler.setFormatter(logging.Formatter(formatStr))
    logger.addHandler(handler)
    _debug("Added %s log handler for '%s'" % ('file' if logFile else 'console', logger.name))

def _configureLogger(name, force=False):
    try:
        logger = _Loggers[name]
    except KeyError:
        # add unknown logger names
        logger = getLogger(name)
        _debug("Added unknown logger name '%s'" % name)

    # If not forcing, skip loggers that already have handlers installed
    if not force and logger.handlers:
        return

    global _LogLevels
    if not _LogLevels:
        setLogLevels(getParam('GCAM.LogLevel') or 'WARN')

    if name in _LogLevels:
        level = _LogLevels[name]
        _debug("Configuring %s, level=%s" % (name, level))
        logger.setLevel(level)
    else:
        parent = logger.parent
        # Handle case of user plugins, which aren't in pygcam package
        # but we'd like them to default to the 'pygcam' loglevel
        if isinstance(parent, logging.RootLogger):
            parent = getLogger(PKGNAME)

        logger.setLevel(parent.level)

    logger.propagate = False

    # flush and remove all handlers
    for handler in logger.handlers:

        if not isinstance(handler, logging.NullHandler):
            handler.flush()
        logger.removeHandler(handler)

    logConsole = getParamAsBoolean('GCAM.LogConsole')
    if logConsole:
        consoleFormat = getParam('GCAM.LogConsoleFormat')
        _addHandler(logger, consoleFormat)

    logFile = getParam('GCAM.LogFile')
    if logFile:
        fileFormat = getParam('GCAM.LogFileFormat')
        _addHandler(logger, fileFormat, logFile=logFile)

    if not logger.handlers:
        logger.addHandler(logger, logging.NullHandler())
        _debug("Added NullHandler to root logger")


def configureLogs(force=False):
    '''
    Do basicConfig setup and configure package loggers based on the information
    in the config instance given. Unless force == True, loggers with handlers will
    not be reconfigured.

    :param force: (bool) if True, reconfigure the logs even if already configured.
    :return: none
    '''
    if not configLoaded() or not _Loggers:
        return

    # First configure explicitly named modules
    explicit = _LogLevels.keys()
    for name in explicit:
        _configureLogger(name, force=force)

    # Next do all the implicit ones, setting their log levels
    # to their parent log levels and setting propagate to False.
    for name in _Loggers.keys():
        if name not in explicit:
            _configureLogger(name, force=force)

def setLogLevels(levelStr):
    '''
    Set the logging level string, which can define levels for packages and/or modules.
    Must call configureLogs(force=True) afterwards. Level string can be a single level,
    which is used as the default for all modules, or module-specific settings, e.g.,
    "WARNING, .tool:DEBUG, .utils:INFO, .mcs.util:INFO, my_plugin:DEBUG"

    :param levelStr: (str) comma-delimited module:LEVEL pairs, or just a single LEVEL
    :return: none
    '''
    global _LogLevels
    _LogLevels = parseLevels(levelStr)



'''
.. Created on: 5/11/16
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from collections import defaultdict

from .config import getParam
from .error import PygcamException
from .utils import getBooleanXML, resourceStream
from .XMLFile import XMLFile

#
# Classes to parse queryFiles and the <queries> element of project.xml
# (see pygcam/etc/queries-schema.xsd). These are in a separate file
# for sharing between query.py and project.py
#
class Query(object):
    def __init__(self, node, defaultMap):
        self.name = node.get('name')
        self.saveAs = node.get('saveAs', self.name)
        self.delete = getBooleanXML(node.get('delete', '1'))
        self.useDefault = useDefault = getBooleanXML(node.get('useDefault', '1'))

        # see if the user provided the attribute, or we defaulted to 1
        explicitUseDefault = node.get('useDefault', None) and useDefault

        # Create a list of tuples with (mapName, level) where level may be None
        rewriters = node.findall('rewriter')
        self.rewriters = map(lambda obj: (obj.get('name'), obj.get('level')), rewriters)

        # We add the default map in two cases: (i) user specified some rewriters and explicitly
        # set useDefault="1", or (ii) there are no rewriters and useDefault has not been set to "0".
        if defaultMap and ((rewriters and explicitUseDefault) or (not rewriters and useDefault)):
            self.rewriters.append((defaultMap, None))


class QueryFile(object):
    def __init__(self, node):
        defaultMap = self.defaultMap = node.get('defaultMap', None)

        nodes = node.findall('query')
        self.queries = [Query(node, defaultMap) for node in nodes]

    def queryFilenames(self):
        """
        Return the name used to compose the filename for this query, which
        may differ from the original query name, e.g., if the same query
        needs to be rewritten differently for different purposes.
        """
        names = map(lambda q: q.saveAs, self.queries)
        return names

    @classmethod
    def parse(cls, filename):
        """
        Parse an XML file holding a list of query descriptions.
        :param filename: (str) the name of the XML file to read
        :return: a QueryFile instance.
        """
        xmlFile = XMLFile(filename, schemaPath='etc/queries-schema.xsd', conditionalXML=True)
        return cls(xmlFile.tree.getroot())

#
# Classes to parse rewriteSets.xml (see pygcam/etc/rewriteSets-schema.xsd)
#
class Rewrite(object):
    def __init__(self, node):
        self.From = node.get('from')    # 'from' is a keyword...
        self.to   = node.get('to')
        self.byAEZ = getBooleanXML(node.get('byAEZ', '0'))
        self.byBasin = getBooleanXML(node.get('byBasin', '0'))    # TBD: GCAM5

    def __str__(self):
        # TBD: Add byBasin for GCAM5
        return "<Rewrite from='%s' to='%s' byAEZ='%s'>" % (self.From, self.to, self.byAEZ)


class RewriteSet(object):
    def __init__(self, node):
        self.name  = node.get('name')
        self.level = node.get('level')
        self.byAEZ = getBooleanXML(node.get('byAEZ', '0'))
        self.byBasin = getBooleanXML(node.get('byBasin', '0'))   # TBD: GCAM5
        self.appendValues = getBooleanXML(node.get('append-values', '0'))
        self.rewrites = map(Rewrite, node.findall('rewrite'))

    def __str__(self):
        # TBD: Add byBasin for GCAM5
        return "<RewriteSet name='%s' level='%s' byAEZ='%s' append-values='%s'>" % \
               (self.name, self.level, self.byAEZ, self.appendValues)

    def asRegionMap(self):
        regionMap = defaultdict(list)
        for rewrite in self.rewrites:
            regionMap[rewrite.to].append(rewrite.From)

        return regionMap

class RewriteSetParser(object):
    # store instances by filename to avoid repeated parsing
    cache = {}

    def __init__(self, node, filename):
        rewriteSets = map(RewriteSet, node.findall('rewriteSet'))
        self.rewriteSets = {obj.name : obj for obj in rewriteSets}
        self.filename = filename # for error messages only

    def getRewriteSet(self, name):
        try:
            return self.rewriteSets[name]
        except KeyError:
            raise PygcamException('RewriteSet "%s" not found in file "%s"' % (name, self.filename))

    @classmethod
    def parse(cls, filename=None):
        """
        Parse an XML file holding a list of query result rewrites.
        :param filename: (str) the name of the XML file to read, or, if
           None, the value of config variable GCAM.XmlSetsFile is used.
        :return: a RewriteSetParser instance
        """
        filename = filename or getParam('GCAM.RewriteSetsFile')

        obj = cls.cache.get(filename)
        if obj:
            return obj

        xmlFile = XMLFile(filename, schemaPath='etc/rewriteSets-schema.xsd')
        obj = cls(xmlFile.tree.getroot(), filename)
        cls.cache[filename] = obj
        return obj

    @classmethod
    def getRegionMap(cls, rewriteSetName, filename=None):
        """
        Lookup a RewriteSet in the given file (or in GCAM.RewriteSetsFile)
        and return it as dictionary.

        :param rewriteSetName: (str) the name of the set to look up.
        :param filename: (str or None) path to rewriteSets.xml file, or if
           None, the value of config variable GCAM.RewriteSetsFile is used.
        :return: (dict) a dictionary keyed by aggregate region names,
           with values being a list of standard GCAM region names comprising
           the aggregate.
        """
        rewriteParser = cls.parse()
        rewriteSet = rewriteParser.getRewriteSet(rewriteSetName)
        return rewriteSet.asRegionMap()



'''
.. codeauthor:: Richard Plevin

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
from ..subcommand import SubcommandABC


class MCSCommand(SubcommandABC):

    def __init__(self, subparsers):
        kwargs = {'help' : '''Enable or disable pygcam Monte Carlo Simulation sub-commands.'''}

        super(MCSCommand, self).__init__('mcs', subparsers, kwargs, group='utils', label='MCS')

    def addArgs(self, parser):
        parser.add_argument('mode', choices=['on', 'off', 'status'],
                            help='''Turn MCS mode on or off, or report current setting''')
        return parser

    def run(self, args, tool):
        import os
        from ..config import mcsSentinelFile, usingMCS

        sentinelFile = mcsSentinelFile()

        if args.mode == 'on':
            open(sentinelFile, 'w').close()     # create empty file

        elif args.mode == 'off':
            try:
                os.remove(sentinelFile)         # remove it if it exists
            except:
                pass
        else:
            # Report current mode
            mode = 'on' if usingMCS() else 'off'
            print("MCS mode is %s" % mode)

PluginClass = MCSCommand



#!/usr/bin/env python
"""
.. "compare" sub-command (compares two config files and the files they load)

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016  Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from ..subcommand import SubcommandABC

def driver(args, tool):
    import os
    import subprocess
    from ..config import getParam
    from ..error import FileFormatError
    from ..utils import mkdirs, pathjoin
    from ..XMLFile import XMLFile

    xmlStarlet = getParam('GCAM.XmlStarlet')

    def getFiles(config):
        xmlFile  = XMLFile(config)
        fileElts = xmlFile.tree.findall('//ScenarioComponents/Value')

        files = {}
        for elt in fileElts:
            fileTag = elt.get('name')
            if fileTag in files:
                raise FileFormatError('Config file "%s" has multiple scenario components with tag "%s"' % (config, fileTag))

            files[fileTag] = elt.text

        return files

    def normalizeFiles(name, files, exedir):
        subdir = pathjoin(args.outputDir, name)
        mkdirs(subdir)

        for fileTag, relPath in files.items():
            fileDir = pathjoin(subdir, fileTag)
            mkdirs(fileDir)

            outfile = pathjoin(fileDir, os.path.basename(relPath))
            infile  = pathjoin(exedir, relPath)

            files[fileTag] = outfile

            cmd = "{xmlStarlet} c14n --without-comments '{infile}' | xml fo -s 2 > '{outfile}'"
            cmd = cmd.format(xmlStarlet=xmlStarlet, infile=infile, outfile=outfile)
            print(cmd)
            subprocess.call(cmd, shell=True)

        return files

    files1 = getFiles(args.config1)
    files2 = getFiles(args.config2)

    set1 = set(files1.keys())
    set2 = set(files2.keys())

    if set1 != set2:
        if len(set1 - set2):
            print("Config2 is missing: %s" % list(set1 - set2))
        if len(set2 - set1):
            print("Config1 is missing: %s" % list(set2 - set1))

        return

    print("Config files are the same structurally.")
    files1 = normalizeFiles('config1', files1, args.exedir1)
    files2 = normalizeFiles('config2', files2, args.exedir2)

    diffs = {}
    for fileTag, path1 in files1.items():
        path2 = files2[fileTag]

        cmd = ['diff', path1, path2]
        print(' '.join(cmd))
        if subprocess.call(cmd, shell=False) != 0:
            diffs[fileTag] = (path1, path2)

    if diffs:
        print('The following files files differ:')
        for fileTag, filenames in diffs.items():
            print("%s: %s" % (fileTag, filenames))



class CompareCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Compare two GCAM configuration files and the files they load to
        find differences. Files are compared using "diff" based on matching "name" tags.'''}
        super(CompareCommand, self).__init__('compare', subparsers, kwargs, group='utils')

    def addArgs(self, parser):
        defaultOutputDir = '/tmp/xmlCompare'

        # Positional args
        parser.add_argument('config1',
                            help='''The first of two config files to compare.''')

        parser.add_argument('exedir1',
                            help='''The "exe" from which config1 pathnames should be computed.''')

        parser.add_argument('config2',
                            help='''The second of two config files to compare.''')

        parser.add_argument('exedir2',
                            help='''The "exe" from which config2 pathnames should be computed.''')

        parser.add_argument('-o', '--outputDir', default=defaultOutputDir,
                            help='''The directory in which to create the normalized versions of XML
                            input files for comparison. Default is %s''' % defaultOutputDir)

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



# Copyright (c) 2017  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from pygcam.subcommand import SubcommandABC

class McsSubcommandABC(SubcommandABC):
    def __init__(self, name, subparsers, kwargs, label=None, guiSuppress=False):
        help = kwargs.get('help')
        prefix = '(MCS) '       # add prefix to all MCS subcommands if not already there
        if help and not help.startswith(prefix):
            kwargs['help'] = prefix + help

        super(McsSubcommandABC, self).__init__(name, subparsers, kwargs, group='mcs',
                                               label=None, guiSuppress=guiSuppress)



'''
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
import os
import subprocess as subp
import time
import re

# Deprecated: not needed as of GCAM 4.3. Maybe drop this?

class XvfbException(Exception):
    pass


class Xvfb(object):
    '''
    Class to wrap the Xvfb (X virtual frame buffer) command to acquire a virtual X11 display.
    This allows X11 apps that (think they) need a display to be run in "headless" mode.
    Sets the environment var DISPLAY to the corresponding value (i.e., for display 1, DISPLAY=":1.0")
    '''
    # The first 2 messages in the "|" expression occurs on the Linux systems this has been
    # tested on. The second message occurs on Mac OSX 10.9.
    # This may require updating to run on other systems or versions of these systems.
    lockFailureMsg = "(Could not create server lock file)|(Cannot establish any listening sockets)|(Server is already active for display)"

    def __init__(self, delay=1.0, maxDisplays=20):
        self.delay = delay
        self.maxDisplays = maxDisplays
        self.proc = None
        self.displayNum = None

        self.acquireDisplay()

        # Set environment variable, but save old value to restore later
        self.oldDisplay = os.environ.get("DISPLAY")
        os.environ["DISPLAY"] = ":%d.0" % self.displayNum

    # Defining __enter__ and __exit__ enables use of "with Xvfb(): context manager
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.terminate()

    def terminate(self):
        if self.proc and self.proc.poll() is None:
            self.proc.terminate()

        os.environ["DISPLAY"] = self.oldDisplay or ""


    def acquireDisplay(self):
        '''
        Loop from display number 0 to self.maxDisplays until we've tried them all or
        we succeed in allocating one.
        :return: The integer number of the allocated display.
        '''

        self.displayNum = None
        for displayNum in range(self.maxDisplays):
            display = ":%d" % displayNum

            args = ['Xvfb', display, '-pn', '-audit', '4', '-screen', '0', '800x600x16']

            try:
                self.proc = subp.Popen(args, stdout=subp.PIPE, stderr=subp.STDOUT)

            except Exception, e:
                raise XvfbException(e)

            time.sleep(self.delay)      # allow time for process to start

            if not self.proc:
                raise XvfbException("Xvfb could not be run")

            retcode = self.proc.poll()
            if retcode is None:         # process is still running
                self.displayNum = displayNum
                return displayNum       # we have acquired a display

            # If Xvfb exited, read stdout to see why
            errmsgs = list(self.proc.stdout)
            lockFailure = any(map(lambda line: re.search(self.lockFailureMsg, line), errmsgs))

            if lockFailure:     # display must be allocated to someone else
                continue        # try the next display number

            # Fail if there's any reason other than failing to acquire the lock
            raise XvfbException("Xvfb %s failed: %s" % (display, errmsgs))

        raise XvfbException("Failed to open any display using Xvfb")



import unittest
import os
from glob import glob
import shutil

from pygcam.utils import TempFile, getTempFile

class TestTempFile(unittest.TestCase):
    def setUp(self):
        self.tmpDir = './data/tmp/test'
        self.removeTmpDir()
        os.mkdir(self.tmpDir)

    def tearDown(self):
        self.removeTmpDir()

    def removeTmpDir(self):
        try:
            shutil.rmtree(self.tmpDir)
        except:
            pass

    def test_createTempFile(self):
        file = getTempFile('.txt', tmpDir=self.tmpDir, delete=True)

        with open(file, 'w') as f:
            f.write("Test string\n")

        assert os.path.isfile(file), "File was not created"
        TempFile.remove(file)
        assert not os.path.isfile(file), "File was not deleted"

    def test_deletion(self):
        fileCount = 3
        for i in range(fileCount):
            TempFile(suffix='.xml', tmpDir=self.tmpDir, openFile=True)

        files = glob(os.path.join(self.tmpDir, '*'))
        self.assertEqual(len(files), fileCount)

        TempFile.deleteAll()

        files = glob(os.path.join(self.tmpDir, '*'))
        self.assertEqual(len(files), 0)

    def test_noDeletion(self):
        fileCount = 3
        for i in range(fileCount):
            obj = TempFile(suffix='.xml', tmpDir=self.tmpDir, openFile=True, delete=False)
            os.close(obj.fd)

        files = glob(os.path.join(self.tmpDir, '*'))
        self.assertEqual(len(files), fileCount)

        TempFile.deleteAll()

        files = glob(os.path.join(self.tmpDir, '*'))
        self.assertEqual(len(files), fileCount)
        # print "tmpDir has files:", files


if __name__ == "__main__":
    unittest.main()



'''
.. Created on: 12/1/16
   Map-generation API

.. Copyright (c) 2016 Richard Plevin and University of California
   See the https://opensource.org/licenses/MIT for license details.
'''

from pygcam.diff import readCsv
from pygcam.map import drawMap

filename = '../tests/data/ws/corn-0/queryResults/Purpose-grown_biomass_production-corn-0.csv'
df = readCsv(filename)
imagefile = '/tmp/testmap.png'

regionMap = {
    'United States' : ['USA'],
    'Brazil' : ['Brazil'],
    'Canada' : ['Canada'],
    'Rest of Latin America' : ['Central America and Caribbean',
                               'Mexico', 'Argentina', 'Colombia',
                               'South America_Southern',
                               'South America_Northern']
}

drawMap(df, imagefile, years=(2020, 2050), interpolate=True, sumByYear=True,
        regions=regionMap.keys(), regionMap=regionMap,
        rewriteSet=None, byAEZ=False,
        title='Random Title', palette=None, animate=False)



from lxml import etree as ET
from pygcam.config import getParam
from pygcam.utils import pathjoin
from pygcam.landProtectionUpdate import protect_land

def main():
    import pandas as pd

    ref_ws = getParam('GCAM.RefWorkspace')
    data_sys = pathjoin(ref_ws, 'input', 'gcam-data-system')
    aglu_xml = pathjoin(data_sys, 'xml', 'aglu-xml')

    protected_land_2_xml = pathjoin(aglu_xml, 'protected_land_input_2.xml')
    protected_land_3_xml = pathjoin(aglu_xml, 'protected_land_input_3.xml')

    parser = ET.XMLParser(remove_blank_text=True, remove_comments=True)

    prot2 = ET.parse(protected_land_2_xml, parser)
    prot3 = ET.parse(protected_land_3_xml, parser)

    regionMap = pd.read_csv(pathjoin(data_sys, '_common', 'mappings', 'GCAM_region_names_32reg.csv'), skiprows=3)
    regionMap.set_index('GCAM_region_ID', inplace=True)
    region = regionMap.region

    # Read the CSV defining the region-wide land protection levels for OTAQ project and
    # apply this anew, verifying that the only differences are way into the decimal places
    otaq_prot_file = pathjoin(ref_ws, 'input', 'otaq-modifications', 'aglu', 'OTAQ_land_protection.csv')
    otaq_prot_df = pd.read_csv(otaq_prot_file)

    otaq_prot_df['region'] = otaq_prot_df['GCAM_region_ID'].map(region)
    otaq_prot_df.drop('GCAM_region_ID', axis=1, inplace=True)
    otaq_prot_df.set_index('region', inplace=True)
    fracs = otaq_prot_df.protect_land_fract

    # protected_land_2 has only UnmanagedPasture
    prot2_dict = {reg: [('UnmanagedPasture', fracs[reg])] for reg in fracs.index}
    protect_land(prot2, prot2_dict)

    outfile = "/Users/rjp/Downloads/land_prot2_modified.xml"
    print "Writing", outfile
    prot2.write(outfile, xml_declaration=True, pretty_print=True)

    # protected_land_3 has all the other unmanaged types
    landtypes = ['UnmanagedForest', 'Grassland', 'Shrubland']
    prot3_dict = {reg: [(landtype, fracs[reg]) for landtype in landtypes] for reg in fracs.index}
    protect_land(prot3, prot3_dict)

    outfile = "/Users/rjp/Downloads/land_prot3_modified.xml"
    print "Writing", outfile
    prot3.write(outfile, xml_declaration=True, pretty_print=True)

if __name__ == '__main__':
    main()



import unittest
import os
import subprocess
from pygcam.landProtection import _makeLandClassXpath, _makeRegionXpath, protectLand, runProtectionScenario
from pygcam.windows import IsWindows

class TestLandProtection(unittest.TestCase):
    def setUp(self):
        pass

    def assertFilesEqual(self, outfile, testfile, deleteOnSuccess=True):
        '''
        Run diff on two files and test that there is no difference.
        '''
        # print "diff %s %s" % (outfile, testfile)
        diffCmd = r"c:\cygwin64\bin\diff" if IsWindows else "diff"
        status = subprocess.call([diffCmd, '-w', outfile, testfile], shell=False)  # -w => ignore whitespace
        self.assertEqual(status, 0, 'Files %s and %s differ' % (outfile, testfile))
        if deleteOnSuccess:
            os.unlink(outfile) # doesn't happen if assertion fails, so file can be examined


    def test_makeRegionXpath(self):
        pairs = (('USA', "//region[@name='USA']"),
                 (['EU15', 'Brazil'], "//region[@name='EU15' or @name='Brazil']"),
                 (None, ""))

        for value, expected in pairs:
            xpath = _makeRegionXpath(value)
            self.assertEqual(xpath, expected, 'Expected\n\t%s\ngot\n\t%s' % (expected, xpath))

    def test_makeLandXpath(self):
        pairs = (('Shrubland',
                  './/UnmanagedLandLeaf[starts-with(@name, "Shrubland")]'),
                 (['Shrubland', 'UnmanagedForest'],
                  './/UnmanagedLandLeaf[starts-with(@name, "Shrubland") or starts-with(@name, "UnmanagedForest")]'))

        for value, expected in pairs:
            xpath = _makeLandClassXpath(value)
            self.assertEqual(xpath, expected, 'Expected\n\t%s\ngot\n\t%s' % (expected, xpath))

    def test_createProtected_land_2(self):
        protectedFraction = 0.9
        classes = ['UnmanagedPasture', 'UnmanagedForest', 'Shrubland']
        regions = ['Brazil', 'USA']

        xmlDir = os.path.join('data', 'xml')

        for num in (2, 3):
            infile   = os.path.join(xmlDir, 'partial_land_input_%d.xml' % num)
            outfile  = os.path.join(xmlDir, 'changed_land_input_%d.xml' % num)
            testfile = os.path.join(xmlDir, 'expected_land_input_%d.xml' % num)

            protectLand(infile, outfile, protectedFraction, landClasses=classes, regions=regions)
            self.assertFilesEqual(outfile, testfile)

    def test_protection_scenario(self):
        scenarioName = 'test'
        xmlDir = os.path.join('data', 'xml')
        tmpDir = os.path.join('data', 'tmp')

        scenarioFile = os.path.join(xmlDir, 'protection.xml')

        xmlFiles = map(lambda num: os.path.join(xmlDir, 'partial_land_input_%d.xml' % num), (2, 3))

        runProtectionScenario(scenarioName, outputDir=tmpDir, scenarioFile=scenarioFile,
                              xmlFiles=xmlFiles, inPlace=False)

        for num in (2, 3):
            outfile  = os.path.join(tmpDir, 'partial_land_input_%d.xml' % num)
            testfile = os.path.join(xmlDir, 'test_scenario_land_input_%d.xml' % num)

            self.assertFilesEqual(outfile, testfile)


if __name__ == "__main__":
    unittest.main()
